/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer: Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.ant.jonasbase;

import org.objectweb.jonas.ant.JOnASBaseTask;

/**
 * Allow to configure the WebContainer service
 * @author Florent Benoit
 */
public class WebContainer extends Tasks {

    /**
     * Info for the logger
     */
    private static final String INFO = "[WebContainer] ";

    /**
     * Name of the implementation class for Jetty
     */
    private static final String JETTY_SERVICE = "org.objectweb.jonas.web.jetty50.JettyJWebContainerServiceImpl";

    /**
     * Name of the implementation class for Tomcat
     */
    private static final String TOMCAT_SERVICE = "org.objectweb.jonas.web.wrapper.catalina55.CatalinaJWebContainerServiceWrapper";

    /**
     * Name of the implementation class for the default service (tomcat)
     */
    private static final String DEFAULT_SERVICE = TOMCAT_SERVICE;

    /**
     * Default port number
     */
    private static final String DEFAULT_PORT = "9000";

    /**
     * Name of the service Tomcat
     */
    private static final String TOMCAT_SERVICE_NAME = "<Service name=\"Tomcat-JOnAS\">";

    /**
     * Tomcat Connector token
     */
    private static final String TOMCAT_CONNECTOR_TOKEN = "<!-- Define an AJP";

    /**
     * Flag allowing to know whether cluster service has been enabled
     */
    private boolean clusterIsEnabled = false;

    /**
     * Default listen cluster port number
     */
    private static final String DEFAULT_CLUSTER_LISTEN_PORT = "4001";

    /**
     * Default cluster multicast addr
     */
    private static final String DEFAULT_CLUSTER_MULTICAST_ADDR = "228.0.0.4";

    /**
     * Default cluster multicast port
     */
    private static final String DEFAULT_CLUSTER_MULTICAST_PORT = "45564";

    /**
     * Default constructor
     */
    public WebContainer() {
        super();
    }

    /**
     * Set the port number for the WebContainer
     * @param portNumber the port for the web Container
     */
    public void setPort(String portNumber) {
        // For tomcat
        JReplace propertyReplace = new JReplace();
        propertyReplace.setConfigurationFile(JOnASBaseTask.TOMCAT_CONF_FILE);
        propertyReplace.setToken(DEFAULT_PORT);
        propertyReplace.setValue(portNumber);
        propertyReplace.setLogInfo(INFO + "Setting Tomcat port number to : " + portNumber);
        addTask(propertyReplace);

        // For Jetty
        propertyReplace = new JReplace();
        propertyReplace.setConfigurationFile(JOnASBaseTask.JETTY_CONF_FILE);
        propertyReplace.setToken(DEFAULT_PORT);
        propertyReplace.setValue(portNumber);
        propertyReplace.setLogInfo(INFO + "Setting Jetty port number to : " + portNumber);
        addTask(propertyReplace);
    }


    /**
     * Set the name of the web container : jetty or tomcat
     * @param containerName jetty or tomcat
     */
    public void setName(String containerName) {

        String service = null;

        if (containerName.equalsIgnoreCase("jetty")) {
            service = JETTY_SERVICE;
        } else {
            service = TOMCAT_SERVICE;
        }

        // Replace the service in jonas.properties file
        JReplace propertyReplace = new JReplace();
        propertyReplace.setConfigurationFile(JOnASBaseTask.JONAS_CONF_FILE);
        propertyReplace.setToken(DEFAULT_SERVICE);
        propertyReplace.setValue(service);
        propertyReplace.setLogInfo(INFO + "Setting service to : " + service);
        addTask(propertyReplace);
    }

    /**
     * Set the AJP port number for the WebContainer
     * @param portNumber the AJP port
     */
    public void setAjpPort(String portNumber) {
        // For tomcat
        JReplace propertyReplace = new JReplace();
        propertyReplace.setConfigurationFile(JOnASBaseTask.TOMCAT_CONF_FILE);
        propertyReplace.setToken(TOMCAT_CONNECTOR_TOKEN);
        String value =
        "    <!-- Define an AJP 1.3 Connector on port " + portNumber + " -->" + "\n"
        +
        "    <Connector port=\"" + portNumber + "\" enableLookups=\"false\" redirectPort=\"8443\" protocol=\"AJP/1.3\" />" + "\n"
        +
        TOMCAT_CONNECTOR_TOKEN ;
        propertyReplace.setValue(value);
        propertyReplace.setLogInfo(INFO + "Setting Tomcat AJP Connector to : " + portNumber);
        addTask(propertyReplace);
    }

    /**
     * Set the Director port number for the WebContainer
     * @param portNumber the Director port
     */
    public void setDirectorPort(String portNumber) {
        // For tomcat
        JReplace propertyReplace = new JReplace();
        propertyReplace.setConfigurationFile(JOnASBaseTask.TOMCAT_CONF_FILE);

        propertyReplace.setToken(TOMCAT_CONNECTOR_TOKEN);
        String value =
        "    <!--  Define a Director Connector on port " + portNumber + " -->" + "\n"
        +
        "    <Connector protocol=\"org.enhydra.servlet.connectionMethods.EnhydraDirector.DirectorProtocol\"" + "\n"
        +
        "       port=\"" + portNumber + "\"\n"
        +
        "       threadTimeout = \"300\"" + "\n"
        +
        "       clientTimeout = \"30\"" + "\n"
        +
        "       sessionAffinity = \"false\"" + "\n"
        +
        "       queueSize = \"400\"" + "\n"
        +
        "       numThreads = \"200\"" + "\n"
        +
        "       bindAddress = \"(All Interfaces)\"" + "\n"
        +
        "       authKey = \"(Unauthenticated)\"" + "\n"
        +
        "    />" + "\n"
        +
        TOMCAT_CONNECTOR_TOKEN;

        propertyReplace.setValue(value);
        propertyReplace.setLogInfo(INFO + "Setting Tomcat AJP Connector to : " + portNumber);
        addTask(propertyReplace);
    }

    /**
     * Enable the cluster feature (if not already made)
     */
    private void setCluster() {
        if (!clusterIsEnabled) {
            JReplace propertyReplace = new JReplace();
            propertyReplace.setConfigurationFile(JOnASBaseTask.TOMCAT_CONF_FILE);
            propertyReplace.setToken("</Host>");
            String value = "\n"
                +
                "        <Cluster className=\"org.apache.catalina.cluster.tcp.SimpleTcpCluster\"" + "\n"
                +
                "            managerClassName=\"org.apache.catalina.cluster.session.DeltaManager\"" + "\n"
                +
                "            expireSessionsOnShutdown=\"false\"" + "\n"
                +
                "            useDirtyFlag=\"true\"" + "\n"
                +
                "            notifyListenersOnReplication=\"true\">" + "\n"
                +
                "          <Membership" + "\n"
                +
                "              className=\"org.apache.catalina.cluster.mcast.McastService\"" + "\n"
                +
                "              mcastAddr=\"228.0.0.4\"" + "\n"
                +
                "              mcastPort=\"45564\"" + "\n"
                +
                "              mcastFrequency=\"500\"" + "\n"
                +
                "              mcastDropTime=\"3000\"/>" + "\n"
                +

                "          <Receiver" + "\n"
                +
                "              className=\"org.apache.catalina.cluster.tcp.ReplicationListener\"" + "\n"
                +
                "              tcpListenAddress=\"auto\"" + "\n"
                +
                "              tcpListenPort=\"4001\"" + "\n"
                +
                "              tcpSelectorTimeout=\"100\"" + "\n"
                +
                "              tcpThreadCount=\"6\"/>" + "\n"
                +

                "          <Sender" + "\n"
                +
                "              className=\"org.apache.catalina.cluster.tcp.ReplicationTransmitter\"" + "\n"
                +
                "              replicationMode=\"pooled\"" + "\n"
                +
                "              ackTimeout=\"15000\"/>" + "\n"
                +

                "          <Valve className=\"org.apache.catalina.cluster.tcp.ReplicationValve\"" + "\n"
                +
                "                 filter=\".*\\.gif;.*\\.js;.*\\.jpg;.*\\.png;.*\\.htm;.*\\.html;.*\\.css;.*\\.txt;\"/>" + "\n"
                +

//              "          <Deployer className=\"org.apache.catalina.cluster.deploy.FarmWarDeployer\"" + "\n" +
//              "                    tempDir=\"/tmp/war-temp/\"" + "\n" +
//              "                    deployDir=\"/tmp/war-deploy/\"" + "\n" +
//              "                    watchDir=\"/tmp/war-listen/\"" + "\n" +
//              "                    watchEnabled=\"false\"/>" +"\n" +
                "        </Cluster>" + "\n"
                +
                "      </Host>";
            propertyReplace.setValue(value);
            propertyReplace.setLogInfo(INFO + "Setting Cluster");
            addTask(propertyReplace);
            clusterIsEnabled = true;

        }
    }

    /**
     * Set the listen cluster port number for the WebContainer
     * @param portNumber the Director port
     */
    public void setClusterListenPort(String portNumber) {
        // For tomcat
        setCluster();
        JReplace propertyReplace = new JReplace();
        propertyReplace.setConfigurationFile(JOnASBaseTask.TOMCAT_CONF_FILE);
        propertyReplace.setToken(DEFAULT_CLUSTER_LISTEN_PORT);
        propertyReplace.setValue(portNumber);
        propertyReplace.setLogInfo(INFO + "Setting Cluster listen port : " + portNumber);
        addTask(propertyReplace);
    }
    /**
     * Set the multicast addr for the cluster
     * @param address the multicast address
     */
    public void setClusterMcastAddr(String address) {
        // For tomcat
        setCluster();
        JReplace propertyReplace = new JReplace();
        propertyReplace.setConfigurationFile(JOnASBaseTask.TOMCAT_CONF_FILE);
        propertyReplace.setToken(DEFAULT_CLUSTER_MULTICAST_ADDR);
        propertyReplace.setValue(address);
        propertyReplace.setLogInfo(INFO + "Setting Cluster multicast addr : " + address);
        addTask(propertyReplace);
    }
    /**
     * Set the multicast port for the cluster
     * @param portNumber the multicast port
     */
    public void setClusterMcastPort(String portNumber) {
        // For tomcat
        setCluster();
        JReplace propertyReplace = new JReplace();
        propertyReplace.setConfigurationFile(JOnASBaseTask.TOMCAT_CONF_FILE);
        propertyReplace.setToken(DEFAULT_CLUSTER_MULTICAST_PORT);
        propertyReplace.setValue(portNumber);
        propertyReplace.setLogInfo(INFO + "Setting Cluster multicast port : " + portNumber);
        addTask(propertyReplace);
    }
}