/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2005 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */
package org.objectweb.jonas.jmx;

import java.io.IOException;
import java.lang.reflect.Method;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import javax.management.InstanceNotFoundException;
import javax.management.JMException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanServerConnection;
import javax.management.MBeanServerNotification;
import javax.management.Notification;
import javax.management.NotificationListener;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.objectweb.carol.rmi.util.PortNumber;
import org.objectweb.carol.util.configuration.CarolDefaultValues;
import org.objectweb.carol.util.configuration.ConfigurationRepository;
import org.objectweb.carol.util.configuration.ProtocolConfiguration;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.discovery.DiscEvent;
import org.objectweb.jonas.service.JonasAlreadyStartedException;
import org.objectweb.jonas.service.ServiceException;

import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * JMX Service implementation.
 * Provides specific doStart() and doStop () methods to start /stop
 * JOnAS JMX Service.

 * @author Adriana Danes.
 *
 */

public class JmxServiceImpl extends AbsJmxServiceImpl implements NotificationListener {

    /**
     * MX4J CommonsLogger fully qualified Classname
     */
    private static final String MX4J_COMMONS_LOGGER_CLASSNAME = "mx4j.log.CommonsLogger";

    /**
     * MX4J Log class
     */
    private static final String MX4J_LOG_CLASSNAME = "mx4j.log.Log";

    /**
     *  MX4J Logger class
     */
    private static final String MX4J_LOGGER_CLASS = "mx4j.log.Logger";

    /**
     * The JNDI name of the JOnAS RMI connector allowing to access the
     * MBeanServer remotelly. Use of this connector was needed before
     * the definition of the Connectors in JMX (before JMX 1.2)
     */
    private String rmiConnectorName = null;

    /**
     * @return The JNDI name of the JOnAS RMI connector
     */
    public String getRmiConnectorName() {
        return this.rmiConnectorName;
    }

    /**
     * Connector servers attached to the MBean server.
     * We may have several connector servers only
     * if several communication protocols are used, as defined by the
     * carol configuration.
     */
    private JMXConnectorServer[] connectorServers = null;

    /**
     * The JMXServiceURLs the connector servers are actually listening on.
     */
    private JMXServiceURL[] connectorServerURLs = null;

    // Domain management support
    /**
     * Mapping serverName -> JMX Service URLs published by the server
     */
    private Map managedServersToUrls = null;
    /**
     * Mapping serverName -> JMX Service connector
     */
    private Map managedServersToConnectors = null;
    /**
     * Mapping serverName -> JMX Service connection
     */
    private Map managedServersToConnections = null;

    /**
     * Management loggers
     */
    private static Logger logger = Log.getLogger(Log.JONAS_JMX_PREFIX);
    private static Logger loggerDomain = Log.getLogger(Log.JONAS_DOMAIN_MANAGEMENT_PREFIX);
    /**
     * Init the logger and then use super method.
     * @param ctx configuration for the init method
     * @throws ServiceException if initialization failed
     */
    public void doInit(Context ctx) throws ServiceException {
        // Test if MX4J CommonsLoggger class is present.
        // In this case, redirect MX4J logging to Jakarta Commons Logging.
        Class mx4jCommonsLoggerClass = null;
        try {
            mx4jCommonsLoggerClass = Thread.currentThread().getContextClassLoader().loadClass(MX4J_COMMONS_LOGGER_CLASSNAME);
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Class " + MX4J_COMMONS_LOGGER_CLASSNAME + " founded");
            }
            Object o =  mx4jCommonsLoggerClass.newInstance();



            // Load Log a Logger class
            Class clazz = Thread.currentThread().getContextClassLoader().loadClass(MX4J_LOG_CLASSNAME);
            Class mx4jLoggerClass = Thread.currentThread().getContextClassLoader().loadClass(MX4J_LOGGER_CLASS);


            // Then get method redirectTo
            Method m = clazz.getMethod("redirectTo", new Class[] {mx4jLoggerClass});
            m.invoke(clazz, new Object[] {o});
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "MX4J logging redirected to the Jakarta commons logger");
            }
        } catch (ClassNotFoundException cnfe) {
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Class " + MX4J_COMMONS_LOGGER_CLASSNAME + " not found: " + cnfe);
            }
        } catch (Exception e) {
            if (logger.isLoggable(BasicLevel.WARN)) {
                logger.log(BasicLevel.WARN, "Problem with " + MX4J_COMMONS_LOGGER_CLASSNAME + " instance creation " + e);
            }
        }
        super.doInit(ctx);
    }
    /**
     * Start the Service.
     * Only need to create a RMI connector
     * @exception ServiceException the service could not be started
     */
    public void doStart() throws ServiceException {
        String serverName = getJonasServerName();
        try {
            // Create a RMI Connector for the JMX agent using the JOnAS RMIConnector class
            // --------------------------------------------------------------------------
            RMIConnector rmiConnector = new RMIConnectorImpl(getJmxServer());
            // Register the connector in JNDI
            InitialContext ictx = new InitialContext();
            rmiConnectorName = "RMIConnector_" + serverName;
            ictx.rebind(rmiConnectorName, rmiConnector);
            ictx.close();

            // Create one or more connector servers
            // (cf. JSR 160, JMX Remote 1.0)
            // Create a JMXServiceURL and a JMXConnectorServer per protocol
            // As some protocols does not have associated connectors, the
            // 2 arrays below may have null values
            // ------------------------------------
            int nbProtocols = ConfigurationRepository.getActiveConfigurationsNumber();
            connectorServerURLs = new JMXServiceURL[nbProtocols];
            connectorServers = new JMXConnectorServer[nbProtocols];
            // Determine protocols used by Carol and their configuration
            ProtocolConfiguration[] protocolConfigurations = ConfigurationRepository.getConfigurations();
            String serviceURL = null;
            for (int i = 0; i < protocolConfigurations.length; i++) {
                ProtocolConfiguration protocolConfiguration = protocolConfigurations[i];
                ConfigurationRepository.setCurrentConfiguration(protocolConfiguration);
                String carolProtocol = protocolConfiguration.getName();
                String providerUrl = protocolConfiguration.getProviderURL();
                URI carolURL = new URI(providerUrl);
                String host = carolURL.getHost();
                String port = String.valueOf(carolURL.getPort());
                String scheme = carolURL.getScheme();
                String ictxFactory = protocolConfiguration.getProtocol().getInitialContextFactoryClassName();
                Properties props = new Properties();
                if (scheme.equals("rmi")) {
                    // Treat RMI/JRMP cas
                    String myName = "jrmpconnector_" + serverName;
                    serviceURL = "service:jmx:rmi://" + host;
                    int jrmpExportedPort = 0;
                    // Add port number of exported objects if one is set by carol.
                    String propertyName = CarolDefaultValues.SERVER_JRMP_PORT;
                    Properties p = ConfigurationRepository.getProperties();
                    if (p != null) {
                        jrmpExportedPort = PortNumber.strToint(p.getProperty(propertyName, "0"), propertyName);
                    }
                    if (jrmpExportedPort > 0) {
                        serviceURL += ":" + jrmpExportedPort;
                    }
                    serviceURL += "/jndi/rmi://"
                        + host + ":" + port + "/"
                        + myName;

                    props.put(Context.INITIAL_CONTEXT_FACTORY, ictxFactory);
                    props.put(Context.PROVIDER_URL, providerUrl);
                } else if (scheme.equals("jrmi")) { // Treat JEREMIE case
                    String myName = "jeremieconnector_" + serverName;
                    serviceURL = "service:jmx:rmi://" + host;
                    int jeremieExportedPort = 0;
                    // Add port number of exported objects if one is set by carol.
                    String propertyName = CarolDefaultValues.SERVER_JEREMIE_PORT;
                    Properties p = ConfigurationRepository.getProperties();
                    if (p != null) {
                        jeremieExportedPort = PortNumber.strToint(p.getProperty(propertyName, "0"), propertyName);
                        // Add 1 to this port for jeremie as the JMX object will not use jeremie to bind but JRMP methods.
                        jeremieExportedPort++;
                    }
                    if (jeremieExportedPort > 1024) {
                        serviceURL += ":" + jeremieExportedPort;
                    }
                    serviceURL += "/jndi/jrmi://"
                        + host + ":" + port + "/"
                        + myName;

                    props.put(Context.INITIAL_CONTEXT_FACTORY, ictxFactory);
                    props.put(Context.PROVIDER_URL, providerUrl);
                } else if (scheme.equals("cmi")) {
                    // Treat CMI case
                    String myName = "cmiconnector_" + serverName;
                    serviceURL = "service:jmx:rmi://" + host
                        + "/jndi/cmi://"
                        + host + ":" + port + "/"
                        + myName;
                    props.put(Context.INITIAL_CONTEXT_FACTORY, ictxFactory);
                    props.put(Context.PROVIDER_URL, providerUrl);
                } else if (scheme.equals("iiop")) {
                    // Treat RMI/IIOP case
                    String myName = "iiopconnector_" + serverName;
                    //serviceURL = "service:jmx:iiop://" + host + "/jndi/" + myName;

                    serviceURL = "service:jmx:iiop://" + host
                        + "/jndi/iiop://"
                        + host + ":" + port + "/"
                        + myName;

                    props.put("java.naming.corba.orb", new InitialContext().lookup("java:comp/ORB"));
                } else {
                    // Currently do not create connectors for other protocols.
                    connectorServerURLs[i] = null;
                    continue;
                }

                JMXServiceURL url = new JMXServiceURL(serviceURL);
                // - Cast to Map is required for JDK 1.5
                // - Set the "jmx.remote.jndi.rebind" to ask the JMX impl to do a rebind() rather than a bind() in the registry
                props.put("jmx.remote.jndi.rebind", "true");
                JMXConnectorServer connectorServer = JMXConnectorServerFactory.newJMXConnectorServer(url, (Map) props, null);
                connectorServers[i] = connectorServer;
                // Create the MBean associated to the connector
                String connectorObjectName = "connector_" + carolProtocol;
                ObjectName connectorServerName = JonasObjectName.jmxConnectorServer(scheme, connectorObjectName);
                getJmxServer().registerMBean(connectorServer, connectorServerName);
                // Start the JMXConnectorServer
                try {
                    connectorServer.start();
                    connectorServerURLs[i] = connectorServer.getAddress();
                } catch (IllegalArgumentException e) {
                    throw e;
                }
            }
        } catch (javax.naming.NameAlreadyBoundException ne) {
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Cannot start JMX service " + ne);
            }
            throw new JonasAlreadyStartedException();
        } catch (Exception e) {
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG, "Cannot start JMX service " + e);
            }
            throw new ServiceException("Cannot start JMX service", e);
        }

        // set back default.
        ConfigurationRepository.setCurrentConfiguration(ConfigurationRepository.getDefaultConfiguration());

        // Registers as listener to notifications emitted by MBeanServerDelegate
        // MBeanServerDelegate is the MBean which sends registration/unregistration notifications
        try {
            ObjectName delegate = ObjectName.getInstance("JMImplementation:type=MBeanServerDelegate");
            getJmxServer().addNotificationListener(delegate, this, null, null);
        } catch (JMException me) {
            if (logger.isLoggable(BasicLevel.DEBUG)) {
                logger.log(BasicLevel.DEBUG,
                        "JMX service could not be added as notification listener for MBeanServerNotifications "
                                + "related to the REGISTRATION or UNREGISTRETION of JOnAS management MBeans");
            }
        }
        // Create data structures allowing to manage the servers in the domain
        managedServersToConnectors = new HashMap();
        managedServersToConnections = new HashMap();
        managedServersToUrls = new HashMap();
    }
    /**
     * Stop this service
     */
    public void doStop() {
        try {
            InitialContext ictx = new InitialContext();
            ictx.unbind("RMIConnector_" + getJonasServerName());
            ictx.close();

            // Desactivates the connector server, that is, stops listening for client connections.
            // Calling this method will also close all client connections that were made by this server.
            // Being a potentialy slow operation, we keep it commented for the moment
            for (int i = 0; i < connectorServers.length; i++) {
                connectorServers[i].stop();
            }
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "Cannot Unbind Jmx RMI Connector" + e);
        }
        // Unregister some MBeans
        ObjectName domainOn = J2eeObjectName.J2EEDomain(getDomainName());
        try {
            getJmxServer().unregisterMBean(domainOn);
        } catch (Exception e) {
            logger.log(BasicLevel.INFO, "Cannot unregister JEEDomain MBean:" + domainOn.toString());
        }
        ObjectName serverOn = J2eeObjectName.J2EEServer(getDomainName(), getJonasServerName());
        try {
            getJmxServer().unregisterMBean(serverOn);
        } catch (Exception e) {
            logger.log(BasicLevel.INFO, "Cannot unregister JEEServer MBean:" + serverOn.toString());
        }
        // Remove internal references to the MBeanServer.
        releaseJmxServer();

        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "JMX Service stopped");
        }
    }


    /**
     *
     * @return The actual adresses on which listen the created connector servers
     */
    public JMXServiceURL[] getConnectorServerURLs() {
        return this.connectorServerURLs;
    }

    /**
     * Treat REGISTRATION/UNREGISTRATION MBeanServerNotification generated for "management" MBeans
     * Treat also notifications generated by Enroller MBean
     * @param notification received notification
     * @param handback received handback
     */
    public void handleNotification(Notification notification, Object handback) {
        if (notification instanceof MBeanServerNotification) {
            String type = notification.getType();
            // ObjectName of the MBean that caused the notification
            ObjectName registeredOn = ((MBeanServerNotification) notification).getMBeanName();
            String name = registeredOn.getKeyProperty("name");
            // The names below are defined in JonasObjectName class
            String j2eeType = registeredOn.getKeyProperty("j2eeType");
            if ("discoveryEnroller".equals(name) || "discoveryClient".equals(name)) {
                if (type.equals(MBeanServerNotification.REGISTRATION_NOTIFICATION)) {
                    try {
                        // register myself as listener of notifications emitted by the Enroller or DiscoveryClient or J2EEDomain MBeans
                        getJmxServer().addNotificationListener(registeredOn, this, null, null);
                        if (loggerDomain.isLoggable(BasicLevel.DEBUG)) {
                            loggerDomain.log(BasicLevel.DEBUG, "JmxService registered as listener to notifs emitted by " + registeredOn);
                        }
                    } catch (InstanceNotFoundException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                }
                if (type.equals(MBeanServerNotification.UNREGISTRATION_NOTIFICATION)) {
                    try {
                        // unregister myself as listener of notifications emitted by the Enroller or DiscoveryClient MBean
                        getJmxServer().removeNotificationListener(registeredOn, this);
                        if (loggerDomain.isLoggable(BasicLevel.DEBUG)) {
                            loggerDomain.log(BasicLevel.DEBUG, "J2EEDomain (this) removed listener for notifs emitted by " + registeredOn);
                        }
                    } catch (InstanceNotFoundException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    } catch (ListenerNotFoundException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                }
            }
        } else {
            // Tread discovery notification
            String type = notification.getType();
            String message = notification.getMessage();
            DiscEvent userData = (DiscEvent) notification.getUserData();
            String source = ((ObjectName) notification.getSource()).toString();
            String state = userData.getState();
            if (loggerDomain.isLoggable(BasicLevel.DEBUG)) {
                loggerDomain.log(BasicLevel.DEBUG, "Treat notification:");
                loggerDomain.log(BasicLevel.DEBUG, "- source: " + source);
                loggerDomain.log(BasicLevel.DEBUG, "- type: " + type);
                loggerDomain.log(BasicLevel.DEBUG, "- data: ");
                loggerDomain.log(BasicLevel.DEBUG, "--- state: " + state);
                loggerDomain.log(BasicLevel.DEBUG, "--- serverName : " + userData.getServerName());
                loggerDomain.log(BasicLevel.DEBUG, "--- domainName : " + userData.getDomainName());
                if (userData.getConnectorURL() != null) {
                    String[] urls = userData.getConnectorURL();
                    for (int i = 0; i < urls.length; i++) {
                        loggerDomain.log(BasicLevel.DEBUG, "--- urls : " + urls[i]);
                    }
                }
                loggerDomain.log(BasicLevel.DEBUG, "");
            }
            if (state.equals(DiscEvent.RUNNING)) {
                String[] urls = userData.getConnectorURL();
                for (int i = 0; i < urls.length; i++) {
                    addServer(userData.getDomainName(), userData.getServerName(), urls[i]);
                }
            } else if (state.equals(DiscEvent.STOPPING)) {
                removeServer(userData.getDomainName(), userData.getServerName());
            }
        }
    }
    /**
     * Add a JOnAS server in the management domain in order
     * update the management domain configuration.
     *
     * @param domainName name of the management domain the server belongs to
     * @param serverName server name
     * @param connectorServerURL connector server url
     */
    public void addServer(String domainName, String serverName, String connectorServerURL) {
        if (loggerDomain.isLoggable(BasicLevel.DEBUG)) {
            loggerDomain.log(BasicLevel.DEBUG, "domain name: " + domainName + ", server name: " + serverName + ", url: " + connectorServerURL);
        }
        if (domainName.equals(getDomainName())) {
            // Server that belongs to the current management domain
            if (!managedServersToUrls.containsKey(serverName)) {
                // Unknown server, try to create a connection for it
                boolean created;
                try {
                    created = createConnection(serverName, connectorServerURL);
                } catch (Exception e) {
                    created = false;
                    if (loggerDomain.isLoggable(BasicLevel.WARN)) {
                        loggerDomain.log(BasicLevel.WARN, "Could not create JMX connection for server " + serverName , e);
                    }
                }
                if (created) {
                    ArrayList urls = new ArrayList();
                    urls.add(connectorServerURL);
                    managedServersToUrls.put(serverName, urls);
                }
            } // Else :
            // This server is already known, and has a working connection.
        } else {
            if (loggerDomain.isLoggable(BasicLevel.WARN)) {
                loggerDomain.log(BasicLevel.WARN, "The server named " + serverName + " was not started in the management domain "
                        + getDomainName() + ", but in management domain " + domainName);
            }
        }
    }

    /**
     * Create a MBeanServer connection using the provided url for the provided server
     * @param serverName the name of the server to be connnected to
     * @param connectorServerURL the url used to construct connection address
     * @return true if a connection created
     * @throws NamingException
     * @throws IOException
     */
    private boolean createConnection(String serverName, String connectorServerURL) throws NamingException, IOException {
        boolean created = false;
        MBeanServerConnection connection;
        // create a connector client for the connector server at the given url
        JMXConnector connector = null;
        JMXServiceURL url = new JMXServiceURL(connectorServerURL);
        Map env = null;
        if (url.getProtocol().equals("iiop")) {
            env = new HashMap();
            env.put("java.naming.corba.orb", new InitialContext().lookup("java:comp/ORB"));
        }
        connector = JMXConnectorFactory.newJMXConnector(url, env);
        connector.connect(env);
        connection = connector.getMBeanServerConnection();
        if (connection != null) {
            // OK, connection was established to the connector server.
            testConnection(connection, serverName);
            if (managedServersToConnections.containsKey(serverName)) {
                // A connection already exists.
                // We have the following possibilities:
                // - ignore this connection (close it), so continue to use the old one
                // - replace the old one by the new one (close the old)
                // - keep several connections
                //
                // Currently we keep only one connection, the new one (second approach)
                JMXConnector oldConnector = (JMXConnector) managedServersToConnectors.get(serverName);
                try {
                    oldConnector.close();
                } catch (IOException e1) {
                }
            }
            managedServersToConnections.put(serverName, connection);
            managedServersToConnectors.put(serverName, connector);
            created = true;
            loggerDomain.log(BasicLevel.INFO, "MBeanServerConnection created for connecting to server " + serverName + " using URL: " + connectorServerURL);
        } else {
            if (connector != null) {
                try {
                    connector.close();
                } catch (IOException e1) {
                }
            }
        }
        return created;
    }

    private boolean testConnection(MBeanServerConnection connection, String serverName) {
        boolean ok = false;
        try {
            Integer nb = connection.getMBeanCount();
            /*
            System.out.println("IN TEST connection: got " + nb.toString() + " MBeans in the JMX server of JOnAS server " + serverName);
            */
            ok = true;
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
        return ok;
    }

    /**
     * Add a JOnAS server in the management domain
     * @param domainName name of the management domain the server belongs
     * @param serverName server name
     */
    public void removeServer(String domainName, String serverName) {
        if (domainName.equals(getDomainName())) {
            // Server thet belongs to the current management domain

            managedServersToConnections.remove(serverName);
            JMXConnector connector = (JMXConnector) managedServersToConnectors.remove(serverName);
            if (connector != null) {
            try {
                connector.close();
            } catch (IOException e1) {
            }
            }
            managedServersToUrls.remove(serverName);
        } else {
            if (loggerDomain.isLoggable(BasicLevel.WARN)) {
                loggerDomain.log(BasicLevel.WARN, "The server named " + serverName + " was not started in the management domain "
                        + getDomainName() + ", but in management domain " + domainName);
            }
        }
    }
    /**
     * Return a connection to the JMX serverver of a given JOnAS server
     * @param serverName The name of the server a connection to its JMX server its needed
     * @return A connection to that server's JMX server
     */
    public MBeanServerConnection getServerConnection(String serverName) {
        return (MBeanServerConnection) managedServersToConnections.get(serverName);
    }
}
