/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999-2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.webapp.jonasadmin.deploy;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import javax.management.ObjectName;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.objectweb.jonas.jmx.JonasManagementRepr;
import org.objectweb.jonas.webapp.jonasadmin.JonasAdminJmx;

/**
 * @author Patrick Smith
 * @author Gregory Lapouchnian
 */
public class ApplyDomainDeployConfirmAction extends BaseDeployAction {

    // --------------------------------------------------------- Properties

    private ArrayList servers;

    // --------------------------------------------------------- Public Methods

    public ActionForward executeAction(ActionMapping p_Mapping,
            ActionForm p_Form, HttpServletRequest p_Request,
            HttpServletResponse p_Response) throws IOException,
            ServletException {

        DomainDeployForm oForm = (DomainDeployForm) p_Form;

        // Create a new deployment thread only if there is no other deployment
        // in progress and not after a deployment has been completed.

        // These flags are in place to prevent multiple threads from being created
        // when this action is called again and again by the domainDeployProgress
        // JSP page.
        if (!oForm.getDeploymentInProgress() && !oForm.getDeploymentComplete()) {

            oForm.setDeploymentInProgress(true);

            // Create a new thread to deploy the applications to the servers
            // so that the progress JSP can be refreshed with updated progress.
            new DeployThread(p_Mapping, p_Form, p_Request, p_Response) {

                public void run() {
                    // Form used
                    DomainDeployForm oForm = (DomainDeployForm) p_Form;

                    try {
                        // target servers
                        servers = oForm.getListTargetServers();
                        // apps to be deployed on these servers
                        ArrayList alApps = oForm.getListDeploy();
                        // deploy each app the set of selected servers
                        Map reportList = oForm.getReports();

                        String selectedAction = oForm.getSelectedAction();
                        boolean replaceExisting = oForm.getReplacementOption();

                        for (int i = 0; i < alApps.size(); i++) {
                            String current = (String) alApps.get(i);

                            // attempt to deploy the application, save the report
                            // to display to the user on the progress page and
                            // reset the progress to allow the next deployment
                            // to go through
                            if (selectedAction.equals(DomainDeployForm.DEPLOY)) {
                                if (doDeployOperation(current)) {
                                    reportList.put(current, getReport());
                                    resetProgress();
                                }
                                else {
                                    createException("Another domain deployment operation is running on this server.");
                                }
                            } else if (selectedAction.equals(DomainDeployForm.UPLOAD)) {
                                if (doUploadOperation(current, replaceExisting)) {
                                    reportList.put(current, getReport());
                                    resetProgress();
                                }
                                else {
                                    createException("Another domain deployment operation is running on this server.");
                                }
                            } else if (selectedAction.equals(DomainDeployForm.UPLOADDEPLOY)) {
                                if (doUploadDeployOperation(current, replaceExisting)) {
                                    reportList.put(current, getReport());
                                    resetProgress();
                                }
                                else {
                                    createException("Another domain deployment operation is running on this server.");
                                }
                            }
                        }

                        // there is no longer a deployment in progress
                        oForm.setDeploymentInProgress(false);
                        // this current deployment has been completed
                        oForm.setDeploymentComplete(true);
                    } catch (Throwable t) {
                        oForm.setException(true);
                        addGlobalError(t);
                        saveErrors(p_Request, m_Errors);
                    }
                }
            }.start();
        }

        // Forward to the jsp.
        return (p_Mapping.findForward("Domain Deploy Progress"));
    }

    private void createException(String message) throws OperationInProgressException {
        throw new OperationInProgressException(message);
    }

    /**
     * Deploy the application with name 'appName'
     * @param appName the name of the application
     * @return true if the operation was executed, false if the operation was
     * rejected
     */
    private boolean doDeployOperation(String appName) {
        ObjectName on = JonasAdminJmx.getJ2eeDomainObjectName();

        // store the servers in a String[]
        String[] serverArray = new String[servers.size()];
        for (int i = 0; i < servers.size(); i++) {
            serverArray[i] = (String) servers.get(i);
        }

        // define the paramters and the signature for the invocation
        Object[] params = new Object[] {serverArray, appName};
        String[] sig = new String[] {"[Ljava.lang.String;", "java.lang.String"};

        return ((Boolean) JonasManagementRepr.invoke(on,
                getDomainDeploymentMethodName(), params, sig)).booleanValue();
    }

    /**
     * Upload the application with name 'appName'
     * @param appName the name of the application
     * @param replaceExisting if the upload operation should overwrite an existing file.
     * @return true if the operation was executed, false if the operation was
     * rejected
     */
    private boolean doUploadOperation(String appName, boolean replaceExisting) {
        ObjectName on = JonasAdminJmx.getJ2eeDomainObjectName();

        // store the servers in a String[]
        String[] serverArray = new String[servers.size()];
        for (int i = 0; i < servers.size(); i++) {
            serverArray[i] = (String) servers.get(i);
        }

        // define the paramters and the signature for the invocation
        String opName = "uploadFile";
        Object[] params = new Object[] {serverArray, appName, Boolean.valueOf(replaceExisting)};
        String[] sig = new String[] {"[Ljava.lang.String;", "java.lang.String", "boolean"};

        return ((Boolean) JonasManagementRepr.invoke(on,
                opName, params, sig)).booleanValue();
    }

    /**
     * Upload and redeploy the application with name 'appName'
     * @param appName the application to upload and redeploy
     * @param replaceExisting if the upload operation should overwrite an existing file.
     * @return true if the operation was executed, false if the operation was
     * rejected
     */
    private boolean doUploadDeployOperation(String appName, boolean replaceExisting) {
        ObjectName on = JonasAdminJmx.getJ2eeDomainObjectName();

        // store the servers in a String[]
        String[] serverArray = new String[servers.size()];
        for (int i = 0; i < servers.size(); i++) {
            serverArray[i] = (String) servers.get(i);
        }

        // define the paramters and the signature for the invocation
        String opName = getDomainUploadDeployMethodName();
        Object[] params = new Object[] {serverArray, appName, Boolean.valueOf(replaceExisting)};
        String[] sig = new String[] {"[Ljava.lang.String;", "java.lang.String", "boolean"};

        return ((Boolean) JonasManagementRepr.invoke(on,
                opName, params, sig)).booleanValue();
    }

    /**
     * Invoke the resetProgress method of the J2EEDomain.
     */
    private void resetProgress() {
        ObjectName on = JonasAdminJmx.getJ2eeDomainObjectName();
        Object[] params = new Object[] {};
        String[] sig = new String[] {};

        JonasManagementRepr.invoke(on, "resetProgress", params, sig);
    }

    /**
     * Get the report for the deployment operation from J2EEDomain.
     * @return a HashMap containing the results of the deployment of an application
     * on each of the selected servers.
     */
    private HashMap getReport() {
        ObjectName on = JonasAdminJmx.getJ2eeDomainObjectName();
        HashMap report = (HashMap) JonasManagementRepr.getAttribute(on,
                "progressReport");

        return report;
    }

    /**
     * A custom thread class used for doing the deployment operations.
     */
    protected class DeployThread extends Thread {

        protected ActionMapping p_Mapping;

        protected ActionForm p_Form;

        protected HttpServletRequest p_Request;

        protected HttpServletResponse p_Response;

        public DeployThread(ActionMapping p_Mapping, ActionForm p_Form,
                HttpServletRequest p_Request, HttpServletResponse p_Response) {
            this.p_Mapping = p_Mapping;
            this.p_Form = p_Form;
            this.p_Request = p_Request;
            this.p_Response = p_Response;
        }

    }


}
