/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import javax.ejb.EJBException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;

import org.objectweb.jonas_ejb.deployment.api.EntityDesc;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Container Read Uncommitted lock-policy.
 * Only 1 thread can write. All other threads can read, without
 * waiting for a committed state. There is no wait.
 * @author Philippe Durieux
 */
public class JEntitySwitchCRU extends JEntitySwitch {

    /**
     * unique EntityContext
     */
    protected JEntityContext itContext = null;

    /**
     * empty constructor. Object is initialized via init() because it is
     * implemented differently according to jorm mappers.
     */
    public JEntitySwitchCRU() {
        lockpolicy = EntityDesc.LOCK_CONTAINER_READ_UNCOMMITTED;
    }

    protected void initpolicy(JEntityFactory bf) {
        lazyregister = true;
    }

    protected JEntityContext getContext4Tx(Transaction tx) {
        return itContext;
    }

    protected void setContext4Tx(Transaction tx, JEntityContext ctx) {
        itContext = ctx;
    }

    protected void removeContext4Tx(Transaction tx) {
        itContext = null;
    }

    public void waitmyturn(Transaction tx) {
        // No synchro for this policy.
    }

    /**
     * Map a context and its instance.
     * @param tx - the Transaction object
     * @param bctx - the JEntityContext to bind if not null
     * @param forced - force to take this context. (case of create)
     * @param holdit - increment count to hold it, a release will be called
     *        later.
     * @return JEntityContext actually mapped
     */
    public synchronized JEntityContext mapICtx(Transaction tx, JEntityContext bctx, boolean forced, boolean holdit, boolean notused) {

        waitmyturn(tx);

        // Choose the context to use.
        boolean newtrans = false;
        JEntityContext jec = itContext;
        if (forced) {
            // If the new context is enforced, we must first release the older
            if (jec != null) {
                if (TraceEjb.isDebugContext()) {
                    TraceEjb.context.log(BasicLevel.DEBUG, ident + "new context is enforced!");
                }
                discardContext(tx, false, true);
            }
            jec = bctx;
            itContext = jec;
            jec.initEntityContext(this);
            newtrans = true;
        } else {
            if (jec != null) {
                // Reuse the Context for this transaction.
                // If a context was supplied, release it first.
                if (bctx != null) {
                    if (TraceEjb.isDebugContext()) {
                            TraceEjb.context.log(BasicLevel.DEBUG, ident + "a context was supplied!");
                    }
                    bf.releaseJContext(bctx);
                }
                newtrans = true; // TODO: not always a new tx (if more than 1
                                 // modifying call on this)
                jec.reuseEntityContext(newtrans);
            } else {
                if (bctx != null) {
                    jec = bctx;
                } else {
                    // no Context available : get one from the pool.
                    jec = (JEntityContext) bf.getJContext(this);
                }
                jec.initEntityContext(this);
                jec.activate(true);
                itContext = jec; // after activate
                newtrans = true;
            }
        }

        if (tx == null) {
            if (holdit) {
                countIH++;
                if (TraceEjb.isDebugSynchro()) {
                        TraceEjb.synchro.log(BasicLevel.DEBUG, ident + "mapICtx IH count=" + countIH);
                }
                if (shared && countIH == 1) {
                    // reload state that could have been modified by
                    // transactions.
                    jec.activate(false);
                }
            }
        } else {
            if (holdit) {
                // not really used for CRU.
                countIT++;
            }
        }

        return jec;
    }

    /**
     * try to passivate instances
     * @param passivation false if only store bean state.
     * @return false if instance still in memory (busy for example)
     */
    public boolean passivateIH(boolean passivation) {
        JEntityContext jec = getContext4Tx(null);
        // If already passivated, look if we can destroy the objects
        if (jec == null) {
            if (inactivityTimeout > 0) {
                long diff = System.currentTimeMillis() - timestamp;
                if (diff > inactivityTimeout) {
                    if (TraceEjb.isDebugContext()) {
                        TraceEjb.context.log(BasicLevel.DEBUG, ident + " discard object on timeout");
                    }
                    discardContext(null, true, false);
                }
            }
            return true;
        }
        // passivate this instance if required.
        if (passivation) {
            if (TraceEjb.isDebugContext()) {
                TraceEjb.context.log(BasicLevel.DEBUG, "TODO: passivate: " + jec);
            }
        }
        return true;
    }

    public void endIH() {
        return; // NEVER
    }

    /**
     * This transaction has just modified this instance. (CMP2 only)
     * @param tx transaction
     */
    public synchronized void notifyWriting(Transaction tx, JEntityContext bctx) {
        if (TraceEjb.isDebugSynchro()) {
            TraceEjb.synchro.log(BasicLevel.DEBUG, ident);
        }
        if (writingtx != null) {
            if (!tx.equals(writingtx)) {
                TraceEjb.logger.log(BasicLevel.WARN, "");
                try {
                    tx.setRollbackOnly();
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, "cannot set current transaction RollbackOnly");
                }
                // Since we share the same instance, we must rollback also
                // the current transaction (else data would be corrupted).
                try {
                    writingtx.setRollbackOnly();
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, "cannot set other transaction RollbackOnly");
                }
                // We should also avoid modifying instance, in case writingtx is
                // already committing (setRollbackOnly will fail)
                throw new EJBException("Conflict writing entity bean");
            }
        } else {
            // Instance must be written later.
            registerCtx(tx, bctx);
            writingtx = tx;
        }
    }

    /**
     * @return State of this instance. State values are 0=in-tx, 1=out-tx, 2=idle,
     *         3=passive, 4=removed. we don't synchronize this method to avoid
     *         jadmin blocks
     */
    public int getState() {
        if (itContext != null) {
            if (itContext.isMarkedRemoved()) {
                return 4;
            } else {
                if (runningtx != null) {
                    return 0;
                } else {
                    return 2;
                }
            }
        }
        return 3;
    }

}
