/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import javax.ejb.EJBException;
import javax.ejb.NoSuchObjectLocalException;
import javax.ejb.ObjectNotFoundException;
import javax.ejb.TransactionRolledbackLocalException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;

import org.objectweb.jonas_ejb.deployment.api.EntityDesc;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Container Optimistic lock-policy.
 * Only 1 thread can write. All other threads can read, without
 * waiting for a committed state. There is no wait.
 * @author Philippe Durieux
 */
public class JEntitySwitchCRW extends JEntitySwitch {

    /**
     * unique EntityContext
     */
    protected JEntityContext itContext = null;

    // max time to wait when in a transaction, before checking if
    // transaction has been set rollback only (deadlock detection)
    protected static int maxtime = 5000; // 5 sec.

    /**
     * empty constructor. Object is initialized via init() because it is
     * implemented differently according to jorm mappers.
     */
    public JEntitySwitchCRW() {
        lockpolicy = EntityDesc.LOCK_CONTAINER_READ_WRITE;
    }

    protected void initpolicy(JEntityFactory bf) {
        lazyregister = !bf.isPrefetch() && !shared;
    }

    protected JEntityContext getContext4Tx(Transaction tx) {
        return itContext;
    }

    protected void setContext4Tx(Transaction tx, JEntityContext ctx) {
        itContext = ctx;
    }

    protected void removeContext4Tx(Transaction tx) {
        itContext = null;
    }

    /**
     * Must wait only if writingtx is not null, because we must wait
     * for the writing transaction to finish.
     * In all other cases, don't wait to allow more parallellism.
     */
    public void waitmyturn(Transaction tx) {
        int waitcount = 0;
        while (writingtx != null && (tx == null || !tx.equals(writingtx))) {
            if (TraceEjb.isDebugSynchro())
            	TraceEjb.synchro.log(BasicLevel.DEBUG, ident + "WAIT end Tx " + waitcount);
            // deadlock detection
            blockedtx = tx;
            if (waitcount++ > 1 && bf.isBlocked(writingtx) && bf.isBlocking(tx)) {
                try {
                    tx.setRollbackOnly();
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident
                            + "getICtx IT: unexpected exception setting rollbackonly");
                }
                TraceEjb.logger.log(BasicLevel.WARN, ident + "getICtx IT: transaction rolled back");
                throw new TransactionRolledbackLocalException("possible deadlock");
            }
            waiters++;
            try {
                wait(maxtime);
                if (TraceEjb.isDebugSynchro())
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + "NOTIFIED");
            } catch (InterruptedException e) {
                if (TraceEjb.isDebugSynchro())
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + "INTERRUPTED");
            } catch (Exception e) {
                throw new EJBException("JEntitySwitch synchronization pb", e);
            } finally {
                waiters--;
                blockedtx = null;
            }
            // If transaction has been rolledback or set rollback only, give
            // up.
            if (tx != null) {
                int status = Status.STATUS_ROLLEDBACK;
                try {
                    status = tx.getStatus();
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident
                            + "getICtx IT: unexpected exception getting transaction status");
                }
                switch (status) {
                    case Status.STATUS_MARKED_ROLLBACK:
                    case Status.STATUS_ROLLEDBACK:
                    case Status.STATUS_ROLLING_BACK:
                        TraceEjb.logger.log(BasicLevel.WARN, ident + "getICtx IT: transaction rolled back");
                        throw new TransactionRolledbackLocalException("rollback occured while waiting");
                }
            }
        }
    }

    /**
     * Map a context and its instance.
     * @param tx - the Transaction object
     * @param bctx - the JEntityContext to bind if not null
     * @param forced - force to take this context. (case of create)
     * @param holdit - increment count to hold it, a release will be called
     *        later.
     * @return JEntityContext actually mapped
     */
    public synchronized JEntityContext mapICtx(Transaction tx, JEntityContext bctx, boolean forced, boolean holdit, boolean notused) {

        // DEBUG only
        if (bf == null) {
            TraceEjb.synchro.log(BasicLevel.ERROR, "JEntitySwitch not initialized!");
            throw new EJBException("JEntitySwitch not initialized");
        }
        if (TraceEjb.isDebugSynchro()) {
            TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " tx=" + tx);
        }

        waitmyturn(tx);

        // Choose the context to use.
        boolean newtrans = false;
        JEntityContext jec = itContext;
        if (forced) {
            // If the new context is enforced, we must first release the older
            if (jec != null) {
                if (TraceEjb.isDebugContext()) {
                    TraceEjb.context.log(BasicLevel.DEBUG, ident + "new context is enforced!");
                }
                discardContext(tx, false, true);
            }
            jec = bctx;
            itContext = jec;
            jec.initEntityContext(this);
            newtrans = true;
        } else {
            // First check if bean still exists
            if (isremoved) {
                TraceEjb.logger.log(BasicLevel.WARN, ident + " has been removed.");
                throw new NoSuchObjectLocalException("Try to access a bean previously removed");
            }
            if (jec != null) {
                if (todiscard) {
                    TraceEjb.logger.log(BasicLevel.WARN, ident + " has been discarded.");
                    throw new NoSuchObjectLocalException("Try to access a bean previously discarded");
                }
                // Reuse the Context for this transaction.
                // If a context was supplied, release it first.
                if (bctx != null) {
                    if (TraceEjb.isDebugContext()) {
                        TraceEjb.context.log(BasicLevel.DEBUG, ident + " a context was supplied");
                    }
                    bf.releaseJContext(bctx);
                }
                if (runningtx == null) {    // TODO change this checking
                    newtrans = true;
                }
                jec.reuseEntityContext(newtrans);
            } else {
                if (bctx != null) {
                    jec = bctx;
                } else {
                    // no Context available : get one from the pool.
                    jec = (JEntityContext) bf.getJContext(this);
                }
                jec.initEntityContext(this);
                jec.activate(true);
                itContext = jec; // after activate
                newtrans = true;
            }
        }

        if (tx != null) {
            // Register Context now, except if no new transaction
            if (newtrans && !lazyregister) {
                try {
                    registerCtx(tx, jec);
                    if (TraceEjb.isDebugSynchro()) {
                        TraceEjb.synchro.log(BasicLevel.DEBUG, ident + "mapICtx IT: new tx, registerSynchronization");
                    }
                } catch (IllegalStateException e) {
                    if (TraceEjb.synchro.isLoggable(BasicLevel.WARN)) {
                        TraceEjb.synchro.log(BasicLevel.WARN, ident + "mapICtx IT: not registered!", e);
                    }
                }
            }
            if (holdit) {
                countIT++;
                if (TraceEjb.isDebugSynchro()) {
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + "mapICtx countIT=" + countIT);
                }
            }
        } else {
            if (holdit) {
                countIH++;
                if (TraceEjb.isDebugSynchro()) {
                        TraceEjb.synchro.log(BasicLevel.DEBUG, ident + "mapICtx IH count=" + countIH);
                }
                if (shared && countIH == 1) {
                    // reload state that could have been modified by
                    // transactions.
                    jec.activate(false);
                }
            }
        }

        return jec;
    }


    /**
     * try to passivate instances
     * @param passivation false if only store bean state.
     * @return false if instance still in memory (busy for example)
     */
    public boolean passivateIH(boolean passivation) {

        JEntityContext jec = getContext4Tx(null);

        // If already passivated, look if we can destroy the objects
        if (jec == null) {
            if (inactivityTimeout > 0) {
                long diff = System.currentTimeMillis() - timestamp;
                if (diff > inactivityTimeout) {
                    if (TraceEjb.isDebugContext()) {
                        TraceEjb.context.log(BasicLevel.DEBUG, ident + timestamp);
                    }
                    discardContext(null, true, false);
                }
            }
            return true;
        }


        if (runningtx != null) {
            if (TraceEjb.isDebugSynchro()) TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " used in TX");
            return false;
        }

        if (TraceEjb.isDebugSynchro()) TraceEjb.synchro.log(BasicLevel.DEBUG, ident);

        // passivate this instance if required.
        if (passivation) {
            if (TraceEjb.isDebugContext()) {
                TraceEjb.context.log(BasicLevel.DEBUG, "passivated: " + jec);
            }
            jec.passivate();
            bf.releaseJContext(jec);
            removeContext4Tx(null);
            // passivation done, set the timestamp
            timestamp = System.currentTimeMillis();
            if (TraceEjb.isDebugContext()) {
                TraceEjb.context.log(BasicLevel.DEBUG, ident + timestamp);
            }
        }
        return true;
    }

    public void endIH() {
        TraceEjb.synchro.log(BasicLevel.ERROR, ident);
        return; // NEVER
    }

    /**
     * This transaction has just modified this instance. (CMP2 only)
     * @param tx transaction
     */
    public synchronized void notifyWriting(Transaction tx, JEntityContext bctx) {
        TraceEjb.synchro.log(BasicLevel.DEBUG, ident);
        if (writingtx != null) {
            if (!tx.equals(writingtx)) {
                TraceEjb.logger.log(BasicLevel.WARN, "Conflict with " + writingtx);
                TraceEjb.logger.log(BasicLevel.WARN, "Current Tx is " + tx);
                try {
                    tx.setRollbackOnly();
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, "cannot set current transaction RollbackOnly");
                }
                // Since we share the same instance, we must rollback also
                // the current transaction (else data would be corrupted).
                try {
                    writingtx.setRollbackOnly();
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, "cannot set other transaction RollbackOnly");
                }
                // We should also avoid modifying instance, in case writingtx is
                // already committing (setRollbackOnly will fail)
                throw new EJBException("Conflict writing entity bean");
            } else {
                TraceEjb.synchro.log(BasicLevel.WARN, ident + " should be dirty ?");
                return;
            }
        } else {
            if (lazyregister) {
                if (TraceEjb.isDebugSynchro()) {
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " Register Ctx");
                }
                registerCtx(tx, bctx);
            }
            if (TraceEjb.isDebugSynchro()) {
                TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " Set writingtx");
            }
            writingtx = tx;
        }

        // Wait for exclusive access
        // This policy insures that there is only 1 writer at a time.
        // Cannot check with countIT here, because it is incremented also
        // when called in the same tx. This would trigger deadlocks.
        /*
        while (countIT > 1) {
            if (TraceEjb.isDebugSynchro()) {
                TraceEjb.synchro.log(BasicLevel.DEBUG, ident + "WAIT for exclusive writing");
            }
            waiters++;
            blockedtx = tx;
            try {
                wait(maxtime);
            } catch (InterruptedException e) {
                TraceEjb.synchro.log(BasicLevel.WARN, ident + "INTERRUPTED");
            } catch (Exception e) {
                throw new EJBException("JEntitySwitch synchronization pb", e);
            } finally {
                waiters--;
                blockedtx = null;
            }
        }
        */
    }

    /**
     * @return State of this instance. State values are 0=in-tx, 1=out-tx, 2=idle,
     *         3=passive, 4=removed. we don't synchronize this method to avoid
     *         jadmin blocks
     */
    public int getState() {
        if (itContext != null) {
            if (itContext.isMarkedRemoved()) {
                return 4;
            } else {
                if (runningtx != null) {
                    return 0;
                } else {
                    return 2;
                }
            }
        }
        return 3;
    }

/*
    // -----------------------------------------
    // Temporary added for debugging
    // ------------------------------------------
    public void init(JEntityFactory bf, Object pk) {
        super.init(bf, pk);
        TraceEjb.synchro.log(BasicLevel.DEBUG, ident);
    }
    public synchronized boolean tryBindICtx(Transaction tx, JEntityContext bctx, boolean simple) throws ObjectNotFoundException {
        TraceEjb.synchro.log(BasicLevel.DEBUG, ident);
        return super.tryBindICtx(tx, bctx, simple);
    }
    public JEntityContext getICtx(Transaction tx, JEntityContext newctx) {
        TraceEjb.synchro.log(BasicLevel.DEBUG, ident);
        return super.getICtx(tx, newctx);
    }
    public JEntityContext getICtx(Transaction tx, boolean checkr) {
        TraceEjb.synchro.log(BasicLevel.DEBUG, ident);
        return super.getICtx(tx, checkr);
    }
    public synchronized boolean terminate(Transaction tx) {
        TraceEjb.synchro.log(BasicLevel.WARN, ident);
        return super.terminate(tx);
    }
    public synchronized void releaseICtx(Transaction tx) {
        TraceEjb.synchro.log(BasicLevel.DEBUG, ident);
        super.releaseICtx(tx);
    }
    public synchronized void discardICtx(Transaction tx) {
        TraceEjb.synchro.log(BasicLevel.WARN, ident);
        super.discardICtx(tx);
    }
    public synchronized void txCompleted(Transaction tx, boolean committed) {
        TraceEjb.synchro.log(BasicLevel.WARN, ident);
        super.txCompleted(tx, committed);
    }
    protected void registerCtx(Transaction tx, JEntityContext bctx) {
        TraceEjb.synchro.log(BasicLevel.DEBUG, ident);
        super.registerCtx(tx, bctx);
    }
    protected void discardContext(Transaction tx, boolean forgetpk, boolean pool) {
        TraceEjb.synchro.log(BasicLevel.WARN, ident);
        super.discardContext(tx, forgetpk, pool);
    }
*/
}
