/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.util.HashMap;
import java.util.Iterator;

import javax.ejb.EJBException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;

import org.objectweb.jonas_ejb.deployment.api.EntityDesc;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * DataBase lock policy : 1 instance per transactions.
 * Transaction Isolation is managed by the database.
 * @author Philippe Durieux
 */
public class JEntitySwitchDB extends JEntitySwitch {

    /**
     * EntityContext for non-transacted requests
     */
    protected JEntityContext ihContext = null;

    /**
     * Map of EntityContext's for transactions Key is the Transaction.
     */
    protected HashMap itsContext;

    /**
     * empty constructor. Object is initialized via init() because it is
     * implemented differently according to jorm mappers.
     */
    public JEntitySwitchDB() {
        lockpolicy = EntityDesc.LOCK_DATABASE;
        itsContext = new HashMap();
    }

    protected void initpolicy(JEntityFactory bf) {
        // should be false, because usually this policy implies shared=true.
        lazyregister = !bf.isPrefetch() && !shared;
    }

    /**
     * @param tx The Transaction
     * @return the JEntityContext used for this tx
     */
    protected JEntityContext getContext4Tx(Transaction tx) {
        if (tx == null) {
            return ihContext;
        } else {
            return (JEntityContext) itsContext.get(tx);
        }
    }

    /**
     * @param tx The Transaction
     * @param the JEntityContext used for this tx
     */
    protected void setContext4Tx(Transaction tx, JEntityContext ctx) {
        if (tx == null) {
            ihContext = ctx;
        } else {
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG, ident + " add new Context");
            itsContext.put(tx, ctx);
        }
    }

    /**
     *
     */
    protected void removeContext4Tx(Transaction tx) {
        // free this Context
        if (tx == null) {
            ihContext = null;
        } else {
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG, ident + " get rid of this Context");
            itsContext.remove(tx);
        }
    }

    public void waitmyturn(Transaction tx) {
        // synchro done by database.
    }

    /**
     * try to passivate instances
     * @param passivation false if only store bean state.
     * @return false if instance still in memory (busy for example)
     */
    public boolean passivateIH(boolean passivation) {
        // If already passivated, look if we can destroy the objects
        if (ihContext == null && itsContext.size() == 0) {
            if (inactivityTimeout > 0) {
                long diff = System.currentTimeMillis() - timestamp;
                if (diff > inactivityTimeout) {
                    if (TraceEjb.isDebugContext()) {
                        TraceEjb.context.log(BasicLevel.DEBUG, ident + " discard object on timeout");
                    }
                    discardContext(null, true, false);
                }
            }
            return true;
        }

        // passivate this instance if required.
        if (passivation) {
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG, ident);
            // Instance used when no transaction.
            JEntityContext jec = ihContext;
            if (jec != null) {
                if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, "passivated: " + jec);
                jec.passivate();
                bf.releaseJContext(jec);
                ihContext = null;
            }

            // If passivation done, set the timestamp
            if (itsContext.size() == 0) {
                timestamp = System.currentTimeMillis();
            }

            // Instances used in transaction are already passivated
            // when the transaction is over. (See JEntitySwitch.java)
            if (waiters > 0) {
                if (TraceEjb.isDebugSynchro()) {
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
                }
                notifyAll();
            }
        }
        return true;
    }

    public void endIH() {
        return;
    }

    /**
     * Get a context/instance associated with this transaction Called at each
     * request on the bean (including remove)
     * @param tx - the Transaction object
     * @return the BeanContext
     */
    public JEntityContext getICtx(Transaction tx, boolean checkr) {
        // Don't check reentrance in case of database for now
        // because it doesn't work in multithreading (countIT!)
        return mapICtx(tx, null, false, true, false);
    }

    /**
     * This transaction has just modified this instance. (CMP2 only)
     * @param tx transaction
     */
    public synchronized void notifyWriting(Transaction tx, JEntityContext bctx) {
        if (TraceEjb.isDebugSynchro()) {
            TraceEjb.synchro.log(BasicLevel.DEBUG, ident);
        }
        if (writingtx != null) {
            if (!tx.equals(writingtx)) {
                TraceEjb.logger.log(BasicLevel.DEBUG, bf.getEJBName() + " Conflict: 2 transactions attempt to write the same instance");
                // Don't set rollback only: let the database deal with that.
            }
        } else {
            if (lazyregister) {
                // Instance must be written at commit.
                registerCtx(tx, bctx);
            }
            writingtx = tx;
        }
    }

    /**
     * @return State of this instance. State values are 0=in-tx, 1=out-tx, 2=idle,
     *         3=passive, 4=removed. we don't synchronize this method to avoid
     *         jadmin blocks
     */
    public int getState() {
        if (ihContext != null) {
            if (ihContext.isMarkedRemoved()) {
                return 4;
            } else {
                return 2;
            }
        }
        if (itsContext.size() > 0) {
            return 0;
        }
        return 3;
    }

}
