/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.util.LinkedList;
import java.util.NoSuchElementException;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This class is responsible for swapping bean instances to keep memory usage
 * low enough, according to entity max-cache-size value and to passivate
 * instances on timeout when used outside transactions. There is 1 swapper per
 * Container.
 * @author Philippe Durieux
 */
class Swapper extends Thread {

    /**
     * My container
     */
    private JContainer container;

    /**
     * true if still valid. Will be set to false when the container is unloaded.
     */
    private boolean valid;

    /**
     * true when we must synchronize entity beans
     */
    private boolean syncallbeans;

    /**
     * Used to decide to free objects in memory.
     * The goal is to avoid "out of memory exceptions".
     */
    private int memorylow;

    /**
     * list of BeanFactory used to reduce cache size periodically
     */
    private LinkedList bfList = new LinkedList();

    /**
     * list of BeanFactory used to sync dirty instances
     */
    private LinkedList bfsList = new LinkedList();

    /**
     * Value of the timeout to sync bean instances. default timeout value when
     * no action required = 5 mn.
     */
    private int swapTimeout = 5 * 60;

    /**
     * Period to see if memory must be freed. Default value = every 10 loops.
     */
    private int memoryCount = 10;

    /**
     * constructor
     * @param cont The Container
     */
    public Swapper(JContainer cont) {
        super("JonasSwapper");
        valid = true;
        syncallbeans = false;
        memorylow = 0;
        container = cont;
    }

    /**
     * run method for Thread implementation.
     */
    public void run() {

        while (valid) {
            BeanFactory bf = null;
            BeanFactory bfs = null;
            synchronized (this) {
                if (bfsList.size() == 0 && bfList.size() == 0) {
                    try {
                        wait(swapTimeout * 1000);
                    } catch (InterruptedException e) {
                        TraceEjb.logger.log(BasicLevel.ERROR, getName() + ": swapper interrupted", e);
                    } catch (Exception e) {
                        TraceEjb.logger.log(BasicLevel.ERROR, getName() + ": swapper exception", e);
                    }
                }
                try {
                    bf = (BeanFactory) bfList.removeFirst();
                } catch (NoSuchElementException e) {
                    bf = null;
                }
                try {
                    bfs = (BeanFactory) bfsList.removeFirst();
                } catch (NoSuchElementException e) {
                    bfs = null;
                }
            }
            // This can be done with lock off.
            if (bf != null) {
                if (TraceEjb.isDebugSwapper()) {
                    TraceEjb.swapper.log(BasicLevel.DEBUG, "swapper will try to reduce the cache");
                }
                bf.reduceCache();
                continue;
            }
            if (bfs != null) {
                if (TraceEjb.isDebugSwapper()) {
                    TraceEjb.swapper.log(BasicLevel.DEBUG, "swapper will sync dirty instances");
                }
                bfs.sync();
                continue;
            }
            // We were notified on timeout, probably to sync all beans.
            if (memorylow >= 10) {
                if (TraceEjb.isDebugSwapper()) {
                    TraceEjb.swapper.log(BasicLevel.DEBUG, "swapper will passivate all entity factories");
                }
                container.syncAll(true);
                memorylow = 0;
            } else if (syncallbeans) {
                if (TraceEjb.isDebugSwapper()) {
                    TraceEjb.swapper.log(BasicLevel.DEBUG, "swapper will sync all entity factories - " + memorylow);
                }
                container.syncAll(false);
                memorylow++;
            }
        }
    }

    /**
     * stop the swapper.
     */
    public synchronized void stopIt() {
        valid = false;
        notify();
    }

    /**
     * synchronizes all instances accessed outside transaction.
     * @param bf the Bean Factory
     */
    public synchronized void addBeanFactorySync(BeanFactory bf) {
        if (!bfsList.contains(bf)) {
            if (TraceEjb.isDebugSwapper()) {
                TraceEjb.swapper.log(BasicLevel.DEBUG, "" + bfsList.size());
            }
            bfsList.addLast(bf);
            notify();
        }
    }

    /**
     * add a BeanFactory to the list
     * @param bf the Bean Factory
     */
    public synchronized void addBeanFactory(BeanFactory bf) {
        if (!bfList.contains(bf)) {
            if (TraceEjb.isDebugSwapper()) {
                TraceEjb.swapper.log(BasicLevel.DEBUG, "" + bfList.size());
            }
            bfList.addLast(bf);
            // don't notify because it's better to wait that instances are released
            // before reducing the pool.
            // This is called typically when the pool is empty!
        }
    }

    /**
     * set timeout value
     * @param t in seconds
     */
    public synchronized void setSwapperTimeout(int t) {
        if (TraceEjb.isDebugSwapper()) {
            TraceEjb.swapper.log(BasicLevel.DEBUG, "" + t);
        }
        syncallbeans = true;
        if (t < swapTimeout) {
            swapTimeout = t;
        }
        notify();
    }

    /**
     * return the timeout for this container, in milliseconds. 0 means "no
     * timeout".
     * @return timeout value in seconds.
     */
    public int getSwapperTimeout() {
        if (syncallbeans) {
            return swapTimeout;
        } else {
            return 0;
        }
    }
}
