/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2004 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_lib.genclientstub.generator;

import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.objectweb.carol.util.configuration.ConfigurationRepository;

import org.objectweb.common.Cmd;

import org.objectweb.jonas_lib.I18n;
import org.objectweb.jonas_lib.deployment.api.EjbRefDesc;
import org.objectweb.jonas_lib.genbase.GenBaseException;
import org.objectweb.jonas_lib.genbase.archive.Archive;
import org.objectweb.jonas_lib.genbase.archive.Client;
import org.objectweb.jonas_lib.genbase.archive.EjbJar;
import org.objectweb.jonas_lib.genbase.archive.J2EEArchive;
import org.objectweb.jonas_lib.genbase.archive.WebApp;
import org.objectweb.jonas_lib.genbase.generator.AbsGenerator;
import org.objectweb.jonas_lib.genbase.generator.Config;
import org.objectweb.jonas_lib.genclientstub.ClientStubGenException;
import org.objectweb.jonas_lib.loader.AbsModuleClassLoader;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Generator used to generate Stubs for clients
 * @author Florent Benoit
 */
public class Generator extends AbsGenerator {

    /**
     * i18n
     */
    private static I18n i18n = I18n.getInstance(Generator.class);

    /**
     * EjbRef used for this generator
     */
    private EjbRefDesc ejbRef = null;

    /**
     * Class used if no ejbRef is given
     */
    private String intfStubClassName = null;

    /**
     * Archive used by the modifier
     */
    private Archive archive = null;

    /**
     * Creates a new Generator with the given Config.
     * @param config internal configuration object.
     * @param ejbRef reference to the ejb on which we want create the stub
     * @param intfStubClassName name of the interface
     * @param archive given ejbjar, webapp, ...
     * @throws GenBaseException When sources and target temporary directory
     *         cannot be created
     */
    public Generator(Config config, EjbRefDesc ejbRef, String intfStubClassName, Archive archive) throws GenBaseException {
        super(config);
        this.ejbRef = ejbRef;
        this.intfStubClassName = intfStubClassName;
        this.archive = archive;
    }

    /**
     * Generate stub files.
     * @throws ClientStubGenException When generation fails.
     */
    public void generate() throws ClientStubGenException {

        try {
            //URL[] urls = null;
            String archiveClasspath = null;

            J2EEArchive arch = (J2EEArchive) this.archive;
            AbsModuleClassLoader loader = (AbsModuleClassLoader) arch.getModuleClassloader();

            archiveClasspath = loader.getClasspath();

            // itf class
            String itfClass = null;
            if (ejbRef != null) {
                itfClass = ejbRef.getHome();
            } else {
                itfClass = intfStubClassName;
            }

            // TODO Now try to load the class and check if stub is here

            // Add to the rmic Classpath all the lib and classes of the archive
            String classpath = getConfig().getClasspath() + File.separator + archiveClasspath;

            // Build stub

            // Array of commands
            Map commands = new HashMap();

            // Get current protocol
            String rmiName = ConfigurationRepository.getCurrentConfiguration().getProtocol().getName();

            // Build rmic command for jrmp
            //Cmd jrmpCmd = new Cmd(getConfig().getJavaHomeBin() + getConfig().getNameRmic());
            //jrmpCmd.addArgument("-classpath");
            //jrmpCmd.addArgument(classpath);

//            Cmd jeremieCmd = new Cmd(getConfig().getJavaHomeBin() + getConfig().getNameJava());
//            jeremieCmd.addArgument("-classpath");
//            jeremieCmd.addArgument(classpath);
//            jeremieCmd.addArgument("org.objectweb.jeremie.tools.jrmic.JRMICompiler");
//            jeremieCmd.addArgument("-opt");
//            jeremieCmd.addArgument("-owext");
//            commands.put("jeremie", jeremieCmd);

            if (rmiName.equalsIgnoreCase("iiop")) {
                Cmd iiopCmd = new Cmd(getConfig().getJavaHomeBin() + getConfig().getNameRmic());
                iiopCmd.addArgument("-classpath");
                iiopCmd.addArgument(classpath);
                iiopCmd.addArgument("-iiop");
                iiopCmd.addArgument("-poa");
                iiopCmd.addArgument("-always");
                iiopCmd.addArgument("-keepgenerated");
                iiopCmd.addArgument("-g");
                commands.put("iiop", iiopCmd);
            }


            for (Iterator itCmd = commands.keySet().iterator(); itCmd.hasNext();) {
                String protocol = (String) itCmd.next();
                Cmd cmd = (Cmd) commands.get(protocol);
                cmd.addArgument("-d");
                cmd.addArgument(getClasses().getCanonicalPath());
                cmd.addArguments(getConfig().getJavacOpts());

                // add itf class as arg name
                cmd.addArgument(itfClass);

                if (getLogger().isLoggable(BasicLevel.DEBUG)) {
                    getLogger().log(BasicLevel.DEBUG, "Running '" + cmd.toString() + "'");
                }
                if (cmd.run()) {
                    getLogger().log(BasicLevel.INFO,
                            "Client stubs for classname '" + itfClass + "' successfully generated for protocol '" + protocol + "'.");
                } else {
                    String err = i18n.getMessage("Generator.generate.error");
                    getLogger().log(BasicLevel.ERROR, err);
                    throw new ClientStubGenException(err);
                }
            }

        } catch (Exception e) {
            String err = i18n.getMessage("Generator.generate.error");
            getLogger().log(BasicLevel.ERROR, err);
            throw new ClientStubGenException(err, e);
        }

    }

    /**
     * Compile generated java files into classes directory. Do nothing in case
     * of clientstub generator as all classes are generated
     * @throws ClientStubGenException When compilation fails
     */
    public void compile() throws ClientStubGenException {
    }

    /**
     * Add generated files into an Archive
     * @param archive the archive destination of generated files.
     * @throws ClientStubGenException When files cannot be added in the given
     *         Archive.
     */
    public void addFiles(Archive archive) throws ClientStubGenException {
        if (archive instanceof WebApp) {
            archive.addDirectoryIn("WEB-INF/classes/", getClasses());
        } else if (archive instanceof EjbJar) {
            archive.addDirectory(getClasses());
        } else if (archive instanceof Client) {
            archive.addDirectory(getClasses());
        }
    }

}